from core import BasePipeline
from utils import rando_name
from uuid import uuid4
import pandas as pd

#################################################################################################################################################################
class MinimumSpanningTreePipeline(BasePipeline):
    def __init__(self, name, corr_method = "pearson", client = None):
        params = {"corr_method": corr_method}
        BasePipeline.__init__(self, name, "raw_mst", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class MinimumSpanningTree():
    def __init__(self, dataset = None, name = None, pipeline = None, corr_method = "pearson", client = None):
        self.client  = client
        self.type     = "raw_mst"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, corr_method, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, corr_method = "pearson", client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.corr_method = corr_method
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = MinimumSpanningTreePipeline(pipeline_name, corr_method, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def nodes(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "nodes",
                                          extra_params = {})
        try:
           return response['nodes']
        except:
           raise StandardError(response)

    def rankLinks(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankLinks",
                                          extra_params = {})
        try:
           return response['rankLinks']
        except:
           raise StandardError(response)

    def edges(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "edges",
                                          extra_params = {})
        try:
           return response['edges']
        except:
           raise StandardError(response)

    def rankNodes(self, statistic = "closeness_centrality"):
        extra_params = {"statistic": statistic}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankNodes",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['rankNodes'], index = [statistic]).T.sort(columns = statistic)
        except:
           raise StandardError(response)

    def neighborhood(self, node):
        extra_params = {"node": node}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "neighborhood",
                                          extra_params = extra_params)
        try:
           return response['neighborhood']
        except:
           raise StandardError(response)


#################################################################################################################################################################
class CorrelationFilteredGraphPipeline(BasePipeline):
    def __init__(self, name, K = 3, client = None):
        params = {"K": K}
        BasePipeline.__init__(self, name, "raw_cfg", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class CorrelationFilteredGraph():
    def __init__(self, dataset = None, name = None, pipeline = None, K = 3, client = None):
        self.client  = client
        self.type     = "raw_cfg"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, K, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, K = 3, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.K        = K
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = CorrelationFilteredGraphPipeline(pipeline_name, K, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def nodes(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "nodes",
                                          extra_params = {})
        try:
           return response['nodes']
        except:
           raise StandardError(response)

    def rankLinks(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankLinks",
                                          extra_params = {})
        try:
           return response['rankLinks']
        except:
           raise StandardError(response)

    def edges(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "edges",
                                          extra_params = {})
        try:
           return response['edges']
        except:
           raise StandardError(response)

    def rankNodes(self, statistic = "closeness_centrality"):
        extra_params = {"statistic": statistic}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankNodes",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['rankNodes'], index = [statistic]).T.sort(columns = statistic)
        except:
           raise StandardError(response)

    def neighborhood(self, node):
        extra_params = {"node": node}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "neighborhood",
                                          extra_params = extra_params)
        try:
           return response['neighborhood']
        except:
           raise StandardError(response)

#################################################################################################################################################################
class NeighborNetworkGraphPipeline(BasePipeline):
    def __init__(self, name, K = 3, kernel = "euclidean", client = None):
        params = {"K": K, 
                  "kernel": kernel}
        BasePipeline.__init__(self, name, "raw_knn_net", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class NeighborNetworkGraph():
    def __init__(self, dataset = None, name = None, pipeline = None, K = 3, kernel = "euclidean", client = None):
        self.client  = client
        self.type     = "raw_knn_net"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, K, kernel, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, K = 3, kernel = "euclidean", client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.K        = K
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = NeighborNetworkGraphPipeline(pipeline_name, K, kernel, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def nodes(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "nodes",
                                          extra_params = {})
        try:
           return response['nodes']
        except:
           raise StandardError(response)

    def rankLinks(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankLinks",
                                          extra_params = {})
        try:
           return response['rankLinks']
        except:
           raise StandardError(response)

    def edges(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "edges",
                                          extra_params = {})
        try:
           return response['edges']
        except:
           raise StandardError(response)

    def rankNodes(self, statistic = "closeness_centrality"):
        extra_params = {"statistic": statistic}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankNodes",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['rankNodes'], index = [statistic]).T.sort(columns = statistic)
        except:
           raise StandardError(response)

    def neighborhood(self, node):
        extra_params = {"node": node}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "neighborhood",
                                          extra_params = extra_params)
        try:
           return response['neighborhood']
        except:
           raise StandardError(response)





from core import BasePipeline
from utils import rando_name
from uuid import uuid4
import pandas as pd

###################################################################################################################################################################
class MatrixMinimumSpanningTreePipeline(BasePipeline):
    def __init__(self, name, client = None):
        params = {}
        BasePipeline.__init__(self, name, "matrix_mst", client, params)

    def run(self, model, matrix = None, matrix_name = None, matrix_type = None):
        if matrix is not None:
            return BasePipeline.run(self, matrix_name = matrix.name, matrix_type = matrix.type, model = model)
        else:
            return BasePipeline.run(self, matrix_name = matrix_name, matrix_type = matrix_type, model = model)

class MatrixMinimumSpanningTree():
    def __init__(self, matrix = None, matrix_name = None, matrix_type = None, name = None, pipeline = None, client = None):
        self.client  = client 
        self.type = "matrix_mst"
        if matrix is None and matrix_name is None and matrix_type is None:
            self.__lazy_init__(name)
        else:
            self.__full_init__(matrix, matrix_name, matrix_type, name, pipeline, client)
            
    def __full_init__(self, matrix, matrix_name = None, matrix_type = None, name = None, pipeline = None, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.matrix   = matrix
        self.pipeline = pipeline
        self.matrix_name = matrix_name
        self.matrix_type = matrix_type
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = MatrixMinimumSpanningTreePipeline(name = pipeline_name, 
                                                              client = client)
        self.response = self.pipeline.run(model = self.name,
                                          matrix = self.matrix,
                                          matrix_name = self.matrix_name,
                                          matrix_type = self.matrix_type)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)
        
    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def rankNodes(self, statistic = "closeness_centrality"):
        extra_params = {"statistic": statistic}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankNodes",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['rankNodes'], index = [statistic]).T.sort(columns = statistic)
        except:
           raise StandardError(response)

    def rankLinks(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankLinks",
                                          extra_params = {})
        try:
           return response['rankLinks']
        except:
           raise StandardError(response)

    def nodes(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "nodes",
                                          extra_params = {})
        try:
           return response['nodes']
        except:
           raise StandardError(response)

    def edges(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "edges",
                                          extra_params = {})
        try:
           return response['edges']
        except:
           raise StandardError(response)

    def neighborhood(self, node):
        extra_params = {"node": node}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "neighborhood",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['neighborhood'], index = ['weight']).T.sort(columns = "weight")
        except:
           raise StandardError(response)


###################################################################################################################################################################
class MatrixKernelPCAPipeline(BasePipeline):
    def __init__(self, name, D = 2, client = None):
        params = {"D": D}
        BasePipeline.__init__(self, name, "matrix_kpca", client, params)

    def run(self, model, matrix = None, matrix_name = None, matrix_type = None):
        if matrix is not None:
            return BasePipeline.run(self, matrix_name = matrix.name, matrix_type = matrix.type, model = model)
        else:
            return BasePipeline.run(self, matrix_name = matrix_name, matrix_type = matrix_type, model = model)

class MatrixKernelPCA():
    def __init__(self, matrix = None, matrix_name = None, matrix_type = None, name = None, pipeline = None, D = 2, client = None):
        self.client  = client
        self.type = "matrix_kpca"
        if matrix is None and matrix_name is None and matrix_type is None:
            self.__lazy_init__(name)
        else:
            self.__full_init__(matrix, matrix_name, matrix_type, name, pipeline, D, client)

    def __full_init__(self, matrix = None, matrix_name = None, matrix_type = None, name = None, pipeline = None, D = 2, client = None):
        self.D = D
        if name == None:
            name = rando_name()
        self.name     = name
        self.matrix   = matrix
        self.pipeline = pipeline
        self.matrix_name = matrix_name
        self.matrix_type = matrix_type
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = MatrixKernelPCAPipeline(D = D,
                                                    name = pipeline_name,
                                                    client = client)
        self.response = self.pipeline.run(model = self.name,
                                          matrix = self.matrix,
                                          matrix_name = self.matrix_name,
                                          matrix_type = self.matrix_type)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def embedding(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "embedding",
                                          extra_params = {})
        try:
           return pd.DataFrame(response['embedding'], index = self.nodes())
        except:
           raise StandardError(response)

    def nodes(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "nodes",
                                          extra_params = {})
        try:
           return response['nodes']
        except:
           raise StandardError(response)

    def meta(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "get_meta",
                                          extra_params = {})
        try:
           return response['get_meta']
        except:
           raise StandardError(response)


###################################################################################################################################################################
class MatrixAgglomeratorPipeline(BasePipeline):
    def __init__(self, name, alpha = 0.1, client = None):
        params = {"alpha":alpha}
        BasePipeline.__init__(self, name, "matrix_agglomeration", client, params)

    def run(self, label_dataset, model, matrix = None, matrix_name = None, matrix_type = None):
        if matrix is not None:
            return BasePipeline.run(self, dataset = label_dataset, matrix_name = matrix.name, matrix_type = matrix.type, model = model)
        else:
            return BasePipeline.run(self, dataset = label_dataset, matrix_name = matrix_name, matrix_type = matrix_type, model = model)

class MatrixAgglomerator():
    def __init__(self, label_dataset = None, matrix = None, matrix_name = None, matrix_type = None, name = None, pipeline = None, alpha = 0.1, client = None):
        self.client  = client
        self.type     = "matrix_agglomeration"
        if label_dataset is None and matrix is None and matrix_name is None and matrix_type is None:
            self.__lazy_init__(name)
        else:
            self.__full_init__(label_dataset, matrix, matrix_name, matrix_type, name, pipeline, alpha, client)

    def __full_init__(self, label_dataset, matrix = None, matrix_name = None, matrix_type = None, name = None, pipeline = None, alpha = 0.1, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.label_dataset  = label_dataset
        self.pipeline = pipeline
        self.matrix   = matrix
        self.matrix_name = matrix_name
        self.matrix_type = matrix_type
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = MatrixAgglomeratorPipeline(name  = pipeline_name,
                                                       alpha = alpha,
                                                       client = client)
        self.response = self.pipeline.run(label_dataset = self.label_dataset, 
                                          model = self.name,
                                          matrix = matrix,
                                          matrix_name = matrix_name,
                                          matrix_type = matrix_type)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def rankDist(self, target, page = 0):
        # todo: predict class given new point
        extra_params = {"target": target,
                        "page": page}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankDist",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['rankDist'], index = ['distance']).T.sort(columns = "distance")
        except:
           raise StandardError(response)

    def getKeys(self):
        # todo: return r^2 for training model
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "getKeys",
                                          extra_params = {})
        try:
           return response['getKeys']
        except:
           raise StandardError(response)






from core import BasePipeline
from utils import rando_name
from uuid import uuid4
import pandas as pd

#################################################################################################################################################################
class KalmanOLSPipeline(BasePipeline):
    def __init__(self, name, init_alpha = None, init_beta = None, trans_cov = None, obs_cov = None, init_cov = None, optimizations = [], client = None):
        params = {"init_alpha": init_alpha,
                  "init_beta": init_beta,
                  "trans_cov": trans_cov,
                  "obs_cov": obs_cov,
                  "init_cov": init_cov,
                  "optimizations": optimizations}
        BasePipeline.__init__(self, name, "kalman_ols", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class KalmanOLS():
    def __init__(self, dataset = None, name = None, pipeline = None, init_alpha = None, init_beta = None, trans_cov = None, obs_cov = None, init_cov = None, optimizations = [], client = None):
        self.client  = client
        self.type     = "kalman_ols"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, init_alpha, init_beta, trans_cov, obs_cov, init_cov, optimizations, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, init_alpha = None, init_beta = None, trans_cov = None, obs_cov = None, init_cov = None, optimizations = [], client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.init_alpha = init_alpha
        self.init_beta = init_beta
        self.trans_cov = trans_cov
        self.obs_cov   = obs_cov
        self.init_cov  = init_cov
        self.optimizations = optimizations
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = KalmanOLSPipeline(pipeline_name, init_alpha, init_beta, trans_cov, obs_cov, init_cov, optimizations, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def getState(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "getState",
                                          extra_params = {})
        try:
           return response['getState']
        except:
           raise StandardError(response)

    def getTrainingData(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "getTrainingData",
                                          extra_params = {})
        try:
           return response['getTrainingData']
        except:
           raise StandardError(response)

    # update the model with new data and return updated state info
    def update(self, X, Y):
        extra_params = {"Y": Y, "X": X}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "update",
                                          extra_params = extra_params)
        try:
           return response['update']
        except:
           raise StandardError(response)



from connect import ConnectIO, Uploader
from utils import rando_name

class BasePipeline():
    def __init__(self, name, type, client = None, params = {}):
        self.name = name
        self.type = type
        self.params = params
        self.client = client
        # create the pipeline on IO cloud
        self.pipeline = self.client.create_pipeline(name = self.name,
                                                    type = self.type,
                                                    params = self.params)
        #print(self.pipeline)
        if "error" in self.pipeline.keys():
            raise StandardError(self.pipeline["error"])
        elif "errorMessage" in self.pipeline.keys():
            raise StandardError(self.pipeline["errorMessage"])        

    #def run(self, dataset, model, extra_params = {}):
    def run(self, model, type = None, dataset = None, matrix_name = None, matrix_type = None, X = None, Y = None, extra_params = {}):
        # put the dataframe to the cloud
        if dataset is not None:
            dataset_name = rando_name()
            #print("putting dataset")
            #print(dataset_name)
            self.client.put_df(dataset_name, dataset)
        else:
            dataset_name = None
        if X is not None:
            X_name = rando_name()
            self.client.put_df(X_name, X)
        else:
            X_name = None
        if Y is not None:
            Y_name = rando_name()
            self.client.put_df(Y_name, Y)
        else:
            Y_name = None
        response = self.client.run_pipeline(name = self.name,
                                            model = model,
                                            type = self.type,
                                            dataset = dataset_name,
                                            matrix_name = matrix_name,
                                            matrix_type = matrix_type,
                                            X = X_name,
                                            Y = Y_name,
                                            extra_params = extra_params)
        return response

""" 
class BaseModel():
    def __init__(self, name, type, client = None):
        self.name = name
        self.type = type
        self.client = client


"""



from core import BasePipeline
from utils import rando_name
from uuid import uuid4
import pandas as pd

#################################################################################################################################################################
class KernelDensityEstimatorPipeline(BasePipeline):
    def __init__(self, name, bandwidth = "scott", client = None):
        params = {"bandwidth": bandwidth}
        BasePipeline.__init__(self, name, "raw_kde", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class KernelDensityEstimator():
    def __init__(self, dataset = None, name = None, pipeline = None, bandwidth = "scott", client = None):
        self.client  = client
        self.type     = "raw_kde"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, bandwidth, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, bandwidth = "scott", client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.bandwidth = bandwidth
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = KernelDensityEstimatorPipeline(pipeline_name, bandwidth, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    # update the model with new data and return updated state info
    def simulate(self, N = 1):
        extra_params = {"N": N}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "simulate",
                                          extra_params = extra_params)
        try:
           return response['simulate']
        except:
           raise StandardError(response)

    def hypercube(self, lower_bounds, upper_bounds):
        extra_params = {"lower_bounds": lower_bounds, 
                        "upper_bounds": upper_bounds}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "hypercube",
                                          extra_params = extra_params)
        try:
           return response['hypercube']
        except:
           raise StandardError(response)

#################################################################################################################################################################
class BasicA2DPipeline(BasePipeline):
    def __init__(self, name, retrain = True, client = None):
        params = {"retrain": retrain}
        BasePipeline.__init__(self, name, "basic_a2d", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class BasicA2D():
    def __init__(self, dataset = None, name = None, pipeline = None, retrain = True, client = None):
        self.client  = client
        self.type     = "basic_a2d"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, retrain, client)
        else:
            self.__lazy_init__(name)
  
    def __full_init__(self, dataset, name = None, pipeline = None, retrain = True, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.retrain  = retrain
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = BasicA2DPipeline(pipeline_name, retrain, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def getState(self):
        extra_params = {}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "getState",
                                          extra_params = extra_params)
        try:
           return response['getState']
        except:
           raise StandardError(response)

    def score(self, value):
        extra_params = {"value": value}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "score",
                                          extra_params = extra_params)
        try:
           return response['score']
        except:
           raise StandardError(response)

    def update(self, value):
        extra_params = {"value": value}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "update",
                                          extra_params = extra_params)
        try:
           return response['update']
        except:
           raise StandardError(response)

#################################################################################################################################################################
class IsolationForestPipeline(BasePipeline):
    def __init__(self, name, rate = 0.1, n_trees = 100, client = None):
        params = {"rate" : min(0.5, rate), 
                  "n_trees": n_trees}
        BasePipeline.__init__(self, name, "isolation_forest", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class IsolationForest():
    def __init__(self, dataset = None, name = None, pipeline = None, rate = 0.1, n_trees = 100, client = None):
        self.client  = client
        self.type     = "isolation_forest"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, rate, n_trees, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, rate = 0.1, n_trees = 100, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.rate  = rate
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = IsolationForestPipeline(pipeline_name, rate, n_trees, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def training_scores(self):
        extra_params = {}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "training_scores",
                                          extra_params = extra_params)
        try:
           return response['training_scores']
        except:
           raise StandardError(response)

    def score(self, points):
        extra_params = {"points": points}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "score",
                                          extra_params = extra_params)
        try:
           return response['score']
        except:
           raise StandardError(response)




from core import BasePipeline
from utils import rando_name
from uuid import uuid4
import pandas as pd

class DistanceMatrixPipeline(BasePipeline):
    def __init__(self, name, kernel = "euclidean", geodesic = False, K = 5, kernel_params = {}, client = None):
        params = {"k": K,
                  "kernel": kernel,
                  "kernel_params": kernel_params,
                  "geodesic": geodesic}
        BasePipeline.__init__(self, name, "dist_matrix", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class DistanceMatrix():
    def __init__(self, dataset = None, name = None, pipeline = None, K = 5, kernel = "euclidean", geodesic = False, kernel_params = {}, client = None):
        self.client  = client
        self.type     = "dist_matrix" 
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, K, kernel, geodesic, kernel_params, client)
        else:
            self.__lazy_init__(name)
    
    # full construction, i.e. start from zero and create it all...
    def __full_init__(self, dataset, name = None, pipeline = None, K = 5, kernel = "euclidean", geodesic = False, kernel_params = {}, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.kernel   = kernel
        self.kernel_params = kernel_params
        self.K        = K
        self.geodesic = geodesic
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = DistanceMatrixPipeline(name = pipeline_name, 
                                                   K    = K, 
                                                   kernel = kernel,
                                                   geodesic = geodesic,
                                                   kernel_params = kernel_params, 
                                                   client = client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name
        
    def rankDist(self, target, page = 0):
        # todo: predict class given new point
        extra_params = {"target": target,
                        "page": page}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankDist",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['rankDist'], index = ['distance']).T.sort(columns = "distance")
        except:
           raise StandardError(response)

    def getKeys(self):
        # todo: return r^2 for training model
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "getKeys",
                                          extra_params = {})
        try:
           return response['getKeys']
        except:
           raise StandardError(response)



API_STAGE = "development"

#import warnings
#warnings.filterwarnings("ignore")
import pandas as pd
import numpy as np
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
import json
import boto3
from StringIO import StringIO
import datetime as dt
import tempfile

region_bucket_map = {"us-east-1":      "slicematrixio",
                     "us-west-1":      "slicematrixiouswest",
                     "eu-central-1":   "slicematrixioeurope1",
                     "ap-southeast-1": "slicematrixioasia1"}

region_api_map    = {"us-east-1":      "ud7p0wre43",
                     "us-west-1":      "4u0pr4ljei",
                     "eu-central-1":   "pozdfzsgae",
                     "ap-southeast-1": "gbq29whx4l"}

class ConnectIO():
  def __init__(self, api_key, region = "us-east-1"):
    self.api_key  = api_key
    self.region   = region
    self.bucket   = region_bucket_map[region]
    self.api      = region_api_map[region]
    self.uploader = Uploader(api_key, self.region, self.bucket, self.api)

  def put_file(self, name, fpath):
    self.uploader.put_file(name, fpath)

  def put_df(self, name, dataframe):
    self.uploader.put_df(name, dataframe)

  def delete_file(self, key):
    self.uploader.delete_file(key)

  def list_files(self):
    return self.uploader.list_files()  

  def create_pipeline(self, name, type, params = {}):
    # make post request to pipeline/create
    #print("creating the pipeline!")
    #print(name)
    url = 'https://' + self.api + '.execute-api.' + self.region + '.amazonaws.com/' + API_STAGE + '/pipelines/create'
    #print url
    headers = {'x-api-key': self.api_key, 'Content-Type': 'application/json'}
    body    = {'name': name, 'type': type, 'params': params}
    r = requests.post(url, verify = False, headers = headers, data=json.dumps(body))
    return json.loads(r.text)
  
  def run_pipeline(self, name, model, type = None, dataset = None, matrix_name = None, matrix_type = None, X = None, Y = None, extra_params = {}, memory = "large"):
    url = 'https://' + self.api + '.execute-api.' + self.region + '.amazonaws.com/' + API_STAGE + '/pipelines/run'
    headers = {'x-api-key': self.api_key, 'Content-Type': 'application/json'}
    body = {'name': name, 'model': model, 'memory': memory, 'type': type}
    if dataset != None:
        body['dataset']     = dataset
    else:
        body['dataset']     = ""
    if matrix_name != None and matrix_type != None:
        body['matrix_name'] = matrix_name
        body['matrix_type'] = matrix_type
    if X != None and Y != None:
        body['X'] = X
        body['Y'] = Y
    for param_key in extra_params.keys():
      body[param_key] = extra_params[param_key]
    #print body
    r = requests.post(url, verify = False, headers = headers, data=json.dumps(body))
    return json.loads(r.text)

  def call_model(self, model, type, method, extra_params = {}, memory = "large"):
    url = 'https://' + self.api + '.execute-api.' + self.region + '.amazonaws.com/' + API_STAGE + '/models/call'
    headers = {'x-api-key': self.api_key, 'Content-Type': 'application/json'}
    body    = {'model': model, 'type': type, 'function': method, 'extra_params': extra_params, 'memory': memory}
    r = requests.post(url, verify = False, headers = headers, data=json.dumps(body))
    #print r.text
    return json.loads(r.text)#[method]

  #def delete_pipeline(self, name, type):

  #def delete_model(self, model, type):
  
class Uploader():
  def __init__(self, api_key, region, bucket, api):
    self.api_key = api_key
    self.region  = region
    self.bucket  = bucket
    self.api     = api
    self.credentials = None

  def check_creds(self):
    if self.credentials == None:
      return False
    else:
      # check if expiration is still good
      now = dt.datetime.now()
      delta = self.credentials['Expiration'] - now
      if delta.total_seconds() <= 0:
        return False
      else:
        return True

  def allocate(self):
    url = 'https://' + self.api + '.execute-api.' + self.region + '.amazonaws.com/' + API_STAGE + '/datasets/authorize'
    headers = {'x-api-key': self.api_key, 'Content-Type': 'application/json'}
    r = requests.post(url, verify = False, headers = headers)
    self.credentials = json.loads(r.text)
    #print self.credentials
    try:
        self.credentials['Expiration'] = pd.to_datetime(self.credentials['Expiration'])
        self.client = boto3.client('s3',
          aws_access_key_id=self.credentials['AccessKeyId'],
          aws_secret_access_key=self.credentials['SecretAccessKey'],
          aws_session_token=self.credentials['SessionToken'],
        )
    except:
        raise StandardException("problem authenticating...")

  def put_file(self, name, fpath):
    if self.check_creds() == False:
      self.allocate()
    return self.client.upload_file(fpath, self.bucket, "uploads/" + self.api_key + "/" + name)

  def put_df(self, name, dataframe):
    if self.check_creds() == False:
      self.allocate()
    tf = tempfile.NamedTemporaryFile()
    dataframe.to_csv(tf.name)
    response = self.client.upload_file(tf.name, self.bucket, "uploads/" + self.api_key + "/" + name)
    tf.close()
    return response

  def delete_file(self, key):
    if self.check_creds() == False:
      self.allocate()
    key = "uploads/" + self.api_key + "/" + key
    return self.client.delete_object(Bucket=self.bucket, Key = key)
  
  def list_files(self): 
    url = 'https://' + self.api + '.execute-api.' + self.region + '.amazonaws.com/' + API_STAGE + '/datasets'
    headers = {'x-api-key': self.api_key}
    r = requests.get(url, verify = False, headers = headers)
    return json.loads(r.text)




from core import BasePipeline
from utils import rando_name, r_squared
from uuid import uuid4
import pandas as pd
import numpy as np

# Random Forest Regressor pipeline #############################################################################################################################################3
class RFRegressorPipeline(BasePipeline):
    def __init__(self, name, n_trees = 8, client = None):
        params = {"n_trees": n_trees,}
        BasePipeline.__init__(self, name, "raw_rfr", client, params)

    def run(self, X, Y, model):
        return BasePipeline.run(self, X = X, Y = Y, model = model)

# Random Forest Regressor model; if pipeline == None then create a pipeline for use with this model
class RFRegressor():
    def __init__(self, X = None, Y = None, name = None, pipeline = None, n_trees = 8, client = None):
        self.client  = client
        self.type     = "raw_rfr"
        if X is not None and Y is not None:
            self.__full_init__(X, Y, name, pipeline, n_trees, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, X, Y, name = None, pipeline = None, n_trees = 8, client = None):
        self.n_trees = n_trees 
        if name == None:
            name = rando_name()
        # else:
        #    todo: add feature to instantiate RFRegressor just from name
        #    i.e. an already created model
        self.name     = name
        self.X        = X
        self.Y        = Y
        self.pipeline = pipeline
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = RFRegressorPipeline(pipeline_name, n_trees, client)
        self.response = self.pipeline.run(X = self.X, Y = self.Y, model = self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name
        
    def predict(self, point):
        # todo: predict class given new point
        extra_params = {"features": point.values.tolist()}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "predict",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['predict'])
        except:
           raise StandardError(response)

    def score(self):
        Y_hat = self.predict(self.X)
        return r_squared(Y_hat, self.Y)

# K Nearest Neighbors regressor pipeline
class KNNRegressorPipeline(BasePipeline):
    def __init__(self, name, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, client = None):
        params = {"k": K,
                  "kernel": kernel,
                  "algo": algo,
                  "weights": weights,
                  "kernel_params": kernel_params}
        BasePipeline.__init__(self, name, "raw_knn_regressor", client, params)

    def run(self, X, Y, model):
        return BasePipeline.run(self, X = X, Y = Y, model = model)

class KNNRegressor():
    def __init__(self, X = None, Y = None, name = None, pipeline = None, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, client = None):
        self.client  = client
        self.type     = "raw_knn_regressor"
        if X is not None and Y is not None:
            self.__full_init__(X, Y, name, pipeline, K, kernel, algo, weights, kernel_params, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, X, Y, name = None, pipeline = None, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, client = None):
        self.X       = X
        self.Y       = Y
        if name == None:
            name = rando_name()
        self.name     = name
        self.pipeline = pipeline
        self.K        = K
        self.kernel   = kernel
        self.kernel_params = kernel_params
        self.algo     = algo
        self.weights  = weights
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = KNNRegressorPipeline(pipeline_name, K, kernel, algo, weights, kernel_params, client)
        self.response = self.pipeline.run(X = self.X, Y = self.Y, model = self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def predict(self, point):
        extra_params = {"features": point.values.tolist()}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "predict",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['predict'])
        except:
           raise StandardError(response)

    def score(self):
        Y_hat = self.predict(self.X)
        return r_squared(Y_hat, self.Y)


#################################################################################################################################################################################
class KernelRidgeRegressorPipeline(BasePipeline):
    def __init__(self, name, kernel = "linear", alpha = 1.0, kernel_params = {}, client = None):
        params = {"kernel": kernel,
                  "alpha": alpha,
                  "kernel_params": kernel_params}
        BasePipeline.__init__(self, name, "raw_krr", client, params)

    def run(self, X, Y, model):
        return BasePipeline.run(self, X = X, Y = Y, model = model)

class KernelRidgeRegressor():
    def __init__(self, X = None, Y = None, name = None, pipeline = None, kernel = "linear", alpha = 1.0, kernel_params = {}, client = None):
        self.client  = client
        self.type     = "raw_krr"
        if X is not None and Y is not None:
            self.__full_init__(X, Y, name, pipeline, kernel, alpha, kernel_params, client)
        else:
            self.__lazy_init__(name)        

    def __full_init__(self, X, Y, name = None, pipeline = None, kernel = "linear", alpha = 1.0, kernel_params = {}, client = None):
        self.X       = X
        self.Y       = Y
        if name == None:
            name = rando_name()
        self.name     = name
        self.pipeline = pipeline
        self.kernel   = kernel
        self.alpha    = alpha        
        self.kernel_params = kernel_params
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = KernelRidgeRegressorPipeline(pipeline_name, kernel, alpha, kernel_params, client)
        self.response = self.pipeline.run(X = self.X, Y = self.Y, model = self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def predict(self, point):
        extra_params = {"features": point.values.tolist()}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "predict",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['predict'])
        except:
           raise StandardError(response)

    def score(self):
        Y_hat = self.predict(self.X)
        return r_squared(Y_hat, self.Y)




from core import BasePipeline
from utils import rando_name
from uuid import uuid4

# K Nearest Neighbors classifier pipeline
class KNNClassifierPipeline(BasePipeline):
    def __init__(self, name, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, client = None):
        params = {"k": K,
                  "kernel": kernel,
                  "algo": algo,
                  "weights": weights,
                  "kernel_params": kernel_params}
        BasePipeline.__init__(self, name, "raw_knn_classifier", client, params)

    def run(self, dataset, model, class_column):
        return BasePipeline.run(self, dataset = dataset, model = model, extra_params = {"class_column": class_column})

# K Nearest Neighbors classifier model; if pipeline == None then create a pipeline for use with this model
class KNNClassifier():
    def __init__(self, dataset = None, class_column = None, name = None, pipeline = None, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, client = None):
        self.client  = client 
        self.type     = "raw_knn_classifier"
        if dataset is not None:
            self.__full_init__(dataset, class_column, name, pipeline, K, kernel, algo, weights, kernel_params, client)
        else:
            self.__lazy_init__(name)     
   
    def __full_init__(self, dataset, class_column, name = None, pipeline = None, k = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, client = None):
        self.class_column = class_column
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = KNNClassifierPipeline(pipeline_name, k, kernel, algo, weights, kernel_params, client)
        self.response = self.pipeline.run(self.dataset, self.name, self.class_column)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name
        
    def predict(self, point):
        # todo: predict class given new point
        extra_params = {"features": point}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "predict",
                                          extra_params = extra_params)
        try:
           return response['predict']
        except:
           raise StandardError(response)

    def score(self):
        # todo: return r^2 for training model
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "score",
                                          extra_params = {})
        try:
           return response['score']
        except:
           raise StandardError(response)

    def training_preds(self):
        # todo return predictions for training model
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "training_preds",
                                          extra_params = {})
        try:
           return response['training_preds']
        except:
           raise StandardError(response)

    def training_data(self):
        # todo return predictions for training model
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "training_data",
                                          extra_params = {})
        try:
           return response['training_data']
        except:
           raise StandardError(response)

# probalistic neural network classifier pipeline
class PNNClassifierPipeline(BasePipeline):
    def __init__(self, name, sigma = 0.1, client = None):
        params = {"sigma": sigma}
        BasePipeline.__init__(self, name, "raw_pnn", client, params)

    def run(self, dataset, model, class_column):
        return BasePipeline.run(self, dataset = dataset, model = model, extra_params = {"class_column": class_column})

# probalistic neural network classifier model; if pipeline == None then create a pipeline for use with this model
class PNNClassifier():
    def __init__(self, dataset, class_column, name = None, pipeline = None, sigma = 0.1, client = None):
        self.client  = client
        self.class_column = class_column
        self.type     = "raw_pnn"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, sigma, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, sigma = 0.1, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = PNNClassifierPipeline(pipeline_name, sigma, client)
        self.response = self.pipeline.run(self.dataset, self.name, self.class_column)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def predict(self, point):
        # todo: predict class given new point
        extra_params = {"features": point}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "predict",
                                          extra_params = extra_params)
        try:
           return response['predict']
        except:
           raise StandardError(response)

    def score(self):
        # todo: return r^2 for training model
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "score",
                                          extra_params = {})
        try:
           return response['score']
        except:
           raise StandardError(response)

    def training_preds(self):
        # todo return predictions for training model
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "training_preds",
                                          extra_params = {})
        try:
           return response['training_preds']
        except:
           raise StandardError(response)

    def training_data(self):
        # todo return predictions for training model
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "training_data",
                                          extra_params = {})
        try:
           return response['training_data']
        except:
           raise StandardError(response)





from uuid import uuid4
import numpy as np

def rando_name(type = "short"):
    name = str(uuid4())
    if type == "short":
        return name.split("-")[-1]
    else:
        return name.replace("-", "")

def r_squared(Y_hat, Y):
    res = np.sum(np.square(np.subtract(Y.values, Y_hat.values))) #np.sum(np.square(Y.values - Y_hat.values))
    tot = np.sum(np.square(np.subtract(Y.values, np.mean(Y.values))))
    #print res, tot
    return 1. - ( res / tot )



from setuptools import setup

setup(name='slicematrixIO',
      version='0.1.1',
      description='SliceMatrix-IO Python API',
      url='http://www.slicematrix.io',
      author='Hekaton LLC',
      author_email='tynan@slicematrix.com',
      license='GPL',
      packages=['slicematrixIO'],
      install_requires=[
          'numpy>=1',
          'pandas',
          'boto3',
          'requests',
      ],
      zip_safe=False)



from connect import ConnectIO, Uploader

from bayesian_filters import KalmanOLS                                                     # lazy loading enabled
from distributions import KernelDensityEstimator, BasicA2D, IsolationForest                # lazy loading enabled                
from classifiers import KNNClassifier, PNNClassifier                                       # lazy loading enabled
from graphs import MinimumSpanningTree, CorrelationFilteredGraph, NeighborNetworkGraph     # lazy loading enabled 
from matrices import DistanceMatrix                                                        # lazy loading enabled
from matrix_models import MatrixMinimumSpanningTree, MatrixKernelPCA, MatrixAgglomerator   # lazy loading enabled
from manifolds import KernelPCA, LocalLinearEmbedder, LaplacianEigenmapper, Isomap         # lazy loading enabled
from regressors import KNNRegressor, RFRegressor, KernelRidgeRegressor                     # lazy loading enabled

class SliceMatrix():
    def __init__(self, api_key, region = "us-east-1"):
        self.api_key = api_key
        self.client  = ConnectIO(self.api_key, region = region)

    # bayesian filters ##########################################################################################################################################################
    # need to test this
    def KalmanOLS(self, dataset = None, init_alpha = None, init_beta = None, trans_cov = None, obs_cov = None, init_cov = None, optimizations = [], name = None, pipeline = None):
        return KalmanOLS(dataset = dataset,
                         init_alpha = init_alpha,
                         init_beta  = init_beta,
                         trans_cov  = trans_cov,
                         obs_cov  = obs_cov,
                         init_cov = init_cov,
                         optimizations = optimizations,
                         name     = name,
                         pipeline = pipeline,
                         client   = self.client)

    # classifiers ###############################################################################################################################################################
    def KNNClassifier(self, dataset = None, class_column = None, name = None, pipeline = None, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}):
        return KNNClassifier(dataset  = dataset, 
                             class_column = class_column,
                             name     = name, 
                             pipeline = pipeline, 
                             K        = K,
                             kernel   = kernel, 
                             algo     = algo, 
                             weights  = weights, 
                             kernel_params = kernel_params, 
                             client   = self.client)

    def PNNClassifier(self, dataset = None, class_column = None, name = None, pipeline = None, sigma = 0.1):
        return PNNClassifier(dataset  = dataset,
                             class_column = class_column,
                             name     = name,
                             pipeline = pipeline,
                             sigma    = sigma,
                             client   = self.client)

    
    # distributions #############################################################################################################################################################
    def KernelDensityEstimator(self, dataset = None, bandwith = "scott", kernel_params = {}, name = None, pipeline = None):
        return KernelDensityEstimator(dataset  = dataset,
                                      bandwidth = bandwith,
                                      name     = name,
                                      pipeline = pipeline,
                                      client   = self.client)

    def BasicA2D(self, dataset = None, retrain = True, name = None, pipeline = None):
        return BasicA2D(dataset  = dataset,
                        retrain  = retrain,
                        name     = name,
                        pipeline = pipeline,
                        client   = self.client)

    def IsolationForest(self, dataset = None, rate = 0.1, n_trees = 100, name = None, pipeline = None):
        return IsolationForest(dataset  = dataset,
                               rate  = rate,
                               n_trees = n_trees,
                               name     = name,
                               pipeline = pipeline,
                               client   = self.client)

    # regressors ################################################################################################################################################################
    def KNNRegressor(self, X = None, Y = None, K = 5, kernel = "euclidean", algo = "auto", weights = "uniform", kernel_params = {}, name = None, pipeline = None):
        return KNNRegressor(X = X,
                            Y = Y,
                            name     = name,
                            pipeline = pipeline,
                            K        = K,
                            kernel   = kernel,
                            algo     = algo,
                            weights  = weights,
                            kernel_params = kernel_params,
                            client   = self.client)


    def RFRegressor(self, X = None, Y = None, n_trees = 8, name = None, pipeline = None):
        return RFRegressor(X = X, 
                           Y = Y, 
                           n_trees = n_trees, 
                           name = name,
                           pipeline = pipeline,
                           client = self.client)

    def KernelRidgeRegressor(self, X = None, Y = None, kernel = "linear", alpha = 1.0, kernel_params = {}, name = None, pipeline = None):
        return KernelRidgeRegressor(X = X,
                                    Y = Y,
                                    kernel = kernel, 
                                    alpha  = alpha,
                                    kernel_params = kernel_params,
                                    name = name,
                                    pipeline = pipeline,
                                    client = self.client)
         
    # matrices #################################################################################################################################################################
    def DistanceMatrix(self, dataset = None, K = 5, kernel = "euclidean", kernel_params = {}, geodesic = False, name = None, pipeline = None):
        return DistanceMatrix(dataset = dataset,
                              K       = K,
                              kernel  = kernel,
                              kernel_params = kernel_params,
                              geodesic = geodesic,
                              name = name,
                              pipeline = pipeline,
                              client   = self.client)
    
    # matrix models ############################################################################################################################################################
    def MatrixMinimumSpanningTree(self, matrix = None, matrix_name = None, matrix_type = None, name = None, pipeline = None):
        return MatrixMinimumSpanningTree(matrix = matrix,
                                         matrix_name = matrix_name,
                                         matrix_type = matrix_type,
                                         name = name,
                                         pipeline = pipeline,
                                         client   = self.client)

    def MatrixKernelPCA(self, D = 2, matrix = None, matrix_name = None, matrix_type = None, name = None, pipeline = None):
        return MatrixKernelPCA(D = D, 
                               matrix = matrix,
                               matrix_name = matrix_name,
                               matrix_type = matrix_type,
                               name = name,
                               pipeline = pipeline,
                               client   = self.client)

    def MatrixAgglomerator(self, label_dataset = None, alpha = 0.1, matrix = None, matrix_name = None, matrix_type = None, name = None, pipeline = None):
        return MatrixAgglomerator(alpha = alpha, 
                                  label_dataset = label_dataset,
                                  matrix = matrix,
                                  matrix_name = matrix_name,
                                  matrix_type = matrix_type,
                                  name = name,
                                  pipeline = pipeline,
                                  client   = self.client)

    # manifolds ################################################################################################################################################################
    def KernelPCA(self, dataset = None, D = 2, kernel = "linear", alpha = 1.0, invert = False, kernel_params = {}, name = None, pipeline = None):
        return KernelPCA(dataset = dataset,
                         D = D,
                         kernel  = kernel,
                         kernel_params = kernel_params,
                         alpha = alpha,
                         invert = invert,
                         name = name,
                         pipeline = pipeline,
                         client   = self.client)

    def LocalLinearEmbedder(self, dataset = None, D = 2, K = 3, method = "standard", name = None, pipeline = None):                               
        return LocalLinearEmbedder(dataset = dataset,
                                   D = D,
                                   K = K,
                                   method = method, 
                                   name = name, 
                                   pipeline = pipeline,
                                   client   = self.client)

    def LaplacianEigenmapper(self, dataset = None, D = 2, affinity = "knn", K = 5, gamma = 1.0, name = None, pipeline = None):
        return LaplacianEigenmapper(dataset = dataset,
                                    D = D,
                                    K = K,
                                    gamma = gamma,
                                    name = name,
                                    pipeline = pipeline,
                                    client   = self.client)

    def Isomap(self, dataset = None, D = 2, K = 3, name = None, pipeline = None):
        return Isomap(dataset = dataset,
                      D = D,
                      K = K,
                      name = name,
                      pipeline = pipeline,
                      client   = self.client)

    # graphs ##################################################################################################################################################################
    def MinimumSpanningTree(self, dataset = None, corr_method = "pearson", name = None, pipeline = None):
        return MinimumSpanningTree(dataset = dataset,
                                   corr_method = corr_method, 
                                   name = name,
                                   pipeline = pipeline,
                                   client   = self.client)
    
    def CorrelationFilteredGraph(self, dataset = None, K = 3, name = None, pipeline = None):
        return CorrelationFilteredGraph(dataset = dataset,
                                        K = K,
                                        name = name,
                                        pipeline = pipeline,
                                        client   = self.client)

    def NeighborNetworkGraph(self, dataset = None, K = 3, kernel = "euclidean", name = None, pipeline = None):
        return NeighborNetworkGraph(dataset = dataset,
                                    K = K,
                                    kernel = kernel,
                                    name = name,
                                    pipeline = pipeline,
                                    client   = self.client)




from core import BasePipeline
from utils import rando_name
from uuid import uuid4
import pandas as pd

#################################################################################################################################################################
class KernelPCAPipeline(BasePipeline):
    def __init__(self, name, D = 2, kernel = "linear", alpha = 1.0, invert = False, kernel_params = {}, client = None):
        params = {"D": D,
                  "kernel": kernel,
                  "alpha": alpha,
                  "invert": invert,
                  "kernel_params": kernel_params}
        BasePipeline.__init__(self, name, "raw_kpca", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class KernelPCA():
    def __init__(self, dataset = None, name = None, pipeline = None, D = 2, kernel = "linear", alpha = 1.0, invert = False, kernel_params = {}, client = None):
        self.client  = client 
        self.type     = "raw_kpca"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, D, kernel, alpha, invert, kernel_params, client)
        else:
            self.__lazy_init__(name)

    # full construction, i.e. start from zero and create it all...
    def __full_init__(self, dataset, name = None, pipeline = None, D = 2, kernel = "linear", alpha = 1.0, invert = False, kernel_params = {}, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.D        = D
        self.kernel   = kernel
        self.alpha    = alpha
        self.invert   = invert
        self.kernel_params = kernel_params
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = KernelPCAPipeline(pipeline_name, D, kernel, alpha, invert, kernel_params, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)
        
    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def inverse_embedding(self, nodes = True):
        nodes = self.nodes()
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "inverse_embedding",
                                          extra_params = {})
        try:
           return pd.DataFrame(response['inverse_embedding'], index = nodes)
        except:
           raise StandardError(response)

    def embedding(self, nodes = True):
        nodes = self.nodes()
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "embedding",
                                          extra_params = {})
        try:
           return pd.DataFrame(response['embedding'], index = nodes)
        except:
           raise StandardError(response)

    def nodes(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "nodes",
                                          extra_params = {})
        try:
           return response['nodes']
        except:
           raise StandardError(response)

    def meta(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "meta",
                                          extra_params = {})
        try:
           return response['meta']
        except:
           raise StandardError(response)

    def feature_names(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "feature_names",
                                          extra_params = {})
        try:
           return response['feature_names']
        except:
           raise StandardError(response)


#################################################################################################################################################################
class LocalLinearEmbedderPipeline(BasePipeline):
    def __init__(self, name, D = 2, K = 3, method = "standard", client = None):
        params = {"D": D,
                  "k": K,
                  "method": method}
        BasePipeline.__init__(self, name, "raw_lle", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class LocalLinearEmbedder():
    def __init__(self, dataset = None, name = None, pipeline = None, D = 2, K = 3, method = "standard", client = None):
        self.client  = client
        self.type     = "raw_lle"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, D, K, method, client)
        else:
            self.__lazy_init__(name)        

    def __full_init__(self, dataset, name = None, pipeline = None, D = 2, k = 3, method = "standard", client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.D        = D
        self.k        = k
        self.method   = method
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = LocalLinearEmbedderPipeline(pipeline_name, D, k, method, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def embedding(self, nodes = True):
        nodes = self.nodes()
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "embedding",
                                          extra_params = {})
        try:
           return pd.DataFrame(response['embedding'], index = nodes)
        except:
           raise StandardError(response)

    def nodes(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "nodes",
                                          extra_params = {})
        try:
           return response['nodes']
        except:
           raise StandardError(response)

    def recon_error(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "recon_error",
                                          extra_params = {})
        try:
           return response['recon_err']
        except:
           raise StandardError(response)

    def meta(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "meta",
                                          extra_params = {})
        try:
           return response['meta']
        except:
           raise StandardError(response)

    def feature_names(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "feature_names",
                                          extra_params = {})
        try:
           return response['feature_names']
        except:
           raise StandardError(response)


#################################################################################################################################################################
class LaplacianEigenmapperPipeline(BasePipeline):
    def __init__(self, name, D = 2, affinity = "knn", K = 5, gamma = 1.0, client = None):
        params = {"D": D,
                  "K": K,
                  "affinity": affinity,
                  "gamma": gamma}
        BasePipeline.__init__(self, name, "raw_laplacian_eigenmap", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class LaplacianEigenmapper():
    def __init__(self, dataset = None, name = None, pipeline = None, D = 2, affinity = "knn", K = 5, gamma = 1.0, client = None):
        self.client  = client
        self.type     = "raw_laplacian_eigenmap"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, D, affinity, K, gamma, client)
        else:
            self.__lazy_init__(name)        

    def __full_init__(self, dataset, name = None, pipeline = None, D = 2, affinity = "knn", K = 5, gamma = 1.0, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.D        = D
        self.K        = K
        self.affinity = affinity
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = LaplacianEigenmapperPipeline(pipeline_name, D, affinity, K, gamma, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def embedding(self, nodes = True):
        nodes = self.nodes()
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "embedding",
                                          extra_params = {})
        try:
           return pd.DataFrame(response['embedding'], index = nodes)
        except:
           raise StandardError(response)

    def nodes(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "nodes",
                                          extra_params = {})
        try:
           return response['nodes']
        except:
           raise StandardError(response)

    def meta(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "meta",
                                          extra_params = {})
        try:
           return response['meta']
        except:
           raise StandardError(response)

    def feature_names(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "feature_names",
                                          extra_params = {})
        try:
           return response['feature_names']
        except:
           raise StandardError(response)

    def affinity_matrix(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "affinity_matrix",
                                          extra_params = {})
        try:
           return response['affinity_matrix']
        except:
           raise StandardError(response)

#################################################################################################################################################################
class IsomapPipeline(BasePipeline):
    def __init__(self, name, D = 2, K = 3, client = None):
        params = {"D": D,
                  "K": K}
        BasePipeline.__init__(self, name, "raw_isomap", client, params)

    def run(self, dataset, model):
        return BasePipeline.run(self, dataset = dataset, model = model)

class Isomap():
    def __init__(self, dataset, name = None, pipeline = None, D = 2, K = 3, client = None):
        self.client  = client
        self.type     = "raw_isomap"
        if dataset is not None:
            self.__full_init__(dataset, name, pipeline, D, K, client)
        else:
            self.__lazy_init__(name)

    def __full_init__(self, dataset, name = None, pipeline = None, D = 2, K = 3, client = None):
        if name == None:
            name = rando_name()
        self.name     = name
        self.dataset  = dataset
        self.pipeline = pipeline
        self.D        = D
        self.K        = K
        self.type     = "raw_isomap"
        if self.pipeline == None:
            pipeline_name = rando_name()
            self.pipeline = IsomapPipeline(pipeline_name, D, K, client)
        self.response = self.pipeline.run(self.dataset, self.name)
        try:
            # model will be key if success
            model = self.response['model']
            self.name = model.split("/")[-1]
        except:
            # something went wrong creating the model
            raise StandardError(self.response)

    # lazy loading for already persisted models
    def __lazy_init__(self, model_name):
        self.name     = model_name

    def embedding(self, nodes = True):
        nodes = self.nodes()
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "embedding",
                                          extra_params = {})
        try:
           return pd.DataFrame(response['embedding'], index = nodes)
        except:
           raise StandardError(response)

    def nodes(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "nodes",
                                          extra_params = {})
        try:
           return response['nodes']
        except:
           raise StandardError(response)

    def recon_error(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "recon_error",
                                          extra_params = {})
        try:
           return response['recon_error']
        except:
           raise StandardError(response)

    def rankLinks(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankLinks",
                                          extra_params = {})
        try:
           return response['rankLinks']
        except:
           raise StandardError(response)

    def edges(self):
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "edges",
                                          extra_params = {})
        try:
           return response['edges']
        except:
           raise StandardError(response)

    def rankNodes(self, statistic = "closeness_centrality"):
        extra_params = {"statistic": statistic}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "rankNodes",
                                          extra_params = extra_params)
        try:
           return pd.DataFrame(response['rankNodes'], index = [statistic]).T.sort(columns = statistic)
        except:
           raise StandardError(response)

    def neighborhood(self, node):
        extra_params = {"node": node}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "neighborhood",
                                          extra_params = extra_params)
        try:
           return response['neighborhood']
        except:
           raise StandardError(response)

    def search(self, point):
        extra_params = {"point": point}
        response = self.client.call_model(model  = self.name,
                                          type   = self.type,
                                          method = "search",
                                          extra_params = extra_params)
        try:
           return response['search']
        except:
           raise StandardError(response)





